#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2019 Shlomi Fish <shlomif@cpan.org>
#
# Distributed under terms of the Expat license.

import re

from pysol_cards.cards import ms_rearrange
from pysol_cards.random_base import RandomBase

import random


class PysolRandom(RandomBase):
    def setSeed(self, seed):
        seed = self._convertSeed(seed)
        if not 0 <= seed <= self.MAX_SEED:
            raise ValueError("seed is out of range")
        self.seed = seed
        return seed

    def _convertSeed(self, seed):
        return int(seed)

# /***********************************************************************
# // Linear Congruential random generator
# //
# // Knuth, Donald.E., "The Art of Computer Programming,", Vol 2,
# // Seminumerical Algorithms, Third Edition, Addison-Wesley, 1998,
# // p. 106 (line 26) & p. 108
# ************************************************************************/


class _LCBase(PysolRandom):
    def __init__(self, seed=None):
        if seed is None:
            seed = self._getRandomSeed()
        PysolRandom.__init__(self, seed)
        self.initial_seed = seed
        self.setSeed(seed)
        self.origin = self.ORIGIN_UNKNOWN

    def reset(self):
        self.setSeed(self.initial_seed)

    def getSeedAsStr(self):
        return self._getSeedStrPrefix() + str(self.seed)

    def getSeedStr(self):
        return self._getSeedStrPrefix() + str(self.initial_seed)


class LCRandom64(_LCBase):
    MAX_SEED = 0xffffffffffffffff  # 64 bits

    def random(self):
        self.seed = (self.seed * 6364136223846793005 + 1) & self.MAX_SEED
        return ((self.seed >> 21) & 0x7fffffff) / 2147483648.0

    def _getSeedStrPrefix(self):
        return "old"


class LCRandom31(_LCBase):
    MAX_SEED = ((1 << (31 + 2)) - 1)         # 33 bits

    def setSeed(self, seed):
        if not 1 <= seed <= self.MAX_SEED:
            raise ValueError("seed is out of range")
        self.seed = seed
        self.seedx = seed if (seed < 0x100000000) else (seed - 0x100000000)
        return self.seed

    def random(self):
        if (self.seed < 0x100000000):
            ret = self._rand()
            return (ret if (self.seed < 0x80000000) else (ret | 0x8000))
        else:
            return self._randp() + 1

    def _randp(self):
        self.seedx = ((self.seedx) * 214013 + 2531011) & self.MAX_SEED
        return (self.seedx >> 16) & 0xffff

    def _rand(self):
        self.seedx = ((self.seedx) * 214013 + 2531011) & self.MAX_SEED
        return (self.seedx >> 16) & 0x7fff

    def randint(self, a, b):
        return a + self.random() % (b + 1 - a)

    def _getSeedStrPrefix(self):
        return "ms"

    def increaseSeed(self, seed):
        """docstring for increaseSeed"""
        integer_seed = super(LCRandom31, self).increaseSeed(seed)
        return "ms{}".format(integer_seed)

    def str(self, seed):
        if match_ms_deal_prefix(seed) is None:
            return "%05d" % int(seed)
        return seed

    def getstate(self):
        return (
            self.seed,
            self.seedx,
        )

    def setstate(self, token):
        (self.seed, self.seedx,) = token


# * Mersenne Twister random number generator
class MTRandom(RandomBase, random.Random):
    MAX_SEED = 100000000000000000000  # 20 digits

    def __init__(self, seed=None):
        if seed is None:
            seed = self._getRandomSeed()
        RandomBase.__init__(self)
        random.Random.__init__(self, seed)
        self.initial_seed = seed
        self.initial_state = self.getstate()
        self.origin = self.ORIGIN_UNKNOWN

    def getstate(self):
        return (
            RandomBase.getstate(self),
            random.Random.getstate(self)
        )

    def setstate(self, token):
        base, r2 = token
        RandomBase.setstate(self, base)
        random.Random.setstate(self, r2)

    def setSeed(self, seed):
        random.Random.__init__(self, seed)
        self.initial_state = self.getstate()
        return seed

    def reset(self):
        self.setSeed(self.initial_seed)


def shuffle(cards, game_num, which_deals):
    ms = ((game_num <= 32000) or (which_deals == RandomBase.DEALS_MS))
    r = LCRandom31() if ms else MTRandom() if which_deals == \
        RandomBase.DEALS_PYSOLFC else LCRandom64()
    r.setSeed(game_num)
    return r.shuffle(ms_rearrange(cards) if ms else cards)


_ms_pat = re.compile("ms([0-9]+)\\n?\\Z")


def match_ms_deal_prefix(mystring):
    ret = re.match(_ms_pat, mystring)
    return None if not ret else int(ret.group(1))


MS_LONG_BIT = (1 << 1000)
CUSTOM_BIT = (1 << 999)


def random__str2int(s):
    if s == 'Custom':
        return CUSTOM_BIT | MS_LONG_BIT
    m = match_ms_deal_prefix(s)
    if m is not None:
        return (m | MS_LONG_BIT)
    else:
        return int(s)


def random__int2str(myint):
    if ((myint & MS_LONG_BIT) != 0):
        if ((myint & CUSTOM_BIT) != 0):
            return 'Custom'
        return "ms" + str(myint & (~ MS_LONG_BIT))
    else:
        return str(myint)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2025 Shlomi Fish < https://www.shlomifish.org/ >
#
# Licensed under the terms of the MIT license.

"""

        from pysol_cards.cwrap import create_gen
        variant = "freecell"
        DEALS_MS = 0
        gen = create_gen(variant, DEALS_MS)
        deal_idx = 240002
        deal_s = gen(deal_idx)

"""

from pysol_cards.cards import CardRenderer
from pysol_cards.deal_game import Game
from pysol_cards.random import RandomBase


_r = CardRenderer(True)


def create_gen(game_variant, ms):
    which_deals = (
        RandomBase.DEALS_MS if (ms == 0) else RandomBase.DEALS_PYSOLFC
    )
    game = Game(game_variant, 1, which_deals, 13)

    def _ret(deal_idx):
        return game.calc_deal_string(deal_idx, _r)
    return _ret



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2019 Shlomi Fish <shlomif@cpan.org>
#
# Distributed under terms of the Expat license.

"""

"""


class Card(object):
    ACE = 1
    KING = 13

    def __init__(self, id_, rank, suit):
        self.id, self.rank, self.suit = id_, rank, suit
        self.empty, self.flipped = False, False

    def suit_s(self):
        return 'CSHD'[self.suit]

    def is_ace(self):
        return self.rank == self.ACE

    def is_king(self):
        return self.rank == self.KING

    def flip(self, flipped=True):
        ret = Card(self.id, self.rank, self.suit)
        ret.flipped = flipped
        return ret


class CardRenderer(object):
    """docstring for CardRenderer"""
    def __init__(self, print_ts):
        self.print_ts = print_ts

    def to_s(self, card):
        if card.empty:
            return '-'
        ret = self.rank_s(card) + card.suit_s()
        if card.flipped:
            ret = '<' + ret + '>'
        return ret

    def found_s(self, card):
        return card.suit_s() + '-' + self.rank_s(card)

    def rank_s(self, card):
        ret = "0A23456789TJQK"[card.rank]
        if ((not self.print_ts) and ret == 'T'):
            ret = '10'
        return ret

    def render_l(self, lst):
        return [self.to_s(x) for x in lst]

    def l_concat(self, lst):
        return ' '.join(self.render_l(lst))


def createCards(num_decks, max_rank=13):
    ret = []
    id_ = 0
    for _ in range(num_decks):
        for s in range(4):
            for r in range(max_rank):
                ret.append(Card(id_, r + 1, s))
                id_ += 1
    return ret


def ms_rearrange(cards):
    if len(cards) != 52:
        return cards
    c = []
    for i in range(13):
        for j in (0, 39, 26, 13):
            c.append(cards[i + j])
    return c



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2025 Shlomi Fish < https://www.shlomifish.org/ >
#
# Licensed under the terms of the MIT license.

"""
This is a module that implements a command line interface ( CLI )
for generating ranges of deals, which is used by:

https://github.com/shlomif/fc-solve/blob/master/fc-solve/source/board_gen/gen-multiple-pysol-layouts

The interface has not yet stabilized.

"""

import argparse
import os
import sys
from os.path import basename, join
from pathlib import Path

from pysol_cards.cards import CardRenderer
from pysol_cards.deal_game import Game
from pysol_cards.random import RandomBase


class DealsRange:
    def __init__(self, idxs):
        assert idxs.pop(0) == 'seq'
        self.start = int(idxs.pop(0))
        self.end = int(idxs.pop(0))
        assert self.start >= 1
        assert self.end >= self.start

    def run(self):
        return range(self.start, self.end + 1)


class GenMultiParseCmdLine:
    def __init__(self, argv):
        cliargs = argv[1:]
        if cliargs == ["--mode", "shlomif"] or cliargs == ["--mode=shlomif"]:
            self._backend = ShlomifDwimGenDeals()
            return
        parser = argparse.ArgumentParser(
            prog='PROG',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser.add_argument('--concat', action='store_true',
                            help='concatenate the deals into one large file')
        parser.add_argument('--dir', type=str, required=True,
                            help='output dir')
        parser.add_argument('--force', action='store_true',
                            help='override deal files in existing directory')
        parser.add_argument('--game', type=str, default='freecell',
                            help='The Solitaire variant')
        parser.add_argument(
            '--mkdir', action='store_true',
            help='Make the target directory if it does not exist')
        parser.add_argument('--ms', action='store_true',
                            help='MS/FC-Pro Deals')
        parser.add_argument('--prefix', type=str, required=True,
                            help='filename prefix')
        parser.add_argument('--suffix', type=str, required=True,
                            help='filename suffix')
        parser.add_argument('idxs', nargs='+', default=[],
                            help='indexes')
        args = parser.parse_args(argv[1:])
        self._backend = GenMulti(
            concat=args.concat,
            dir_=args.dir,
            force=args.force,
            game_variant=args.game,
            idxs=args.idxs,
            mkdir_=args.mkdir,
            ms=args.ms,
            prefix=args.prefix,
            suffix=args.suffix,
        )

    def run(self):
        return self._backend.run()


class GenMulti:
    def __init__(self, concat, dir_, force, game_variant, idxs,
                 mkdir_, ms, prefix, suffix):
        self.concat = concat
        self.dir_ = dir_
        self.force = force
        self.game_variant = game_variant
        self.idxs = idxs
        self.mkdir_ = mkdir_
        # Sanitize, see:
        # https://stackoverflow.com/questions/6803505
        self.prefix = basename(prefix)
        self.suffix = basename(suffix)
        self.which_deals = (RandomBase.DEALS_MS if ms
                            else RandomBase.DEALS_PYSOLFC)

        self.rend = CardRenderer(True)
        self.game = Game(self.game_variant, 1, self.which_deals, 13)

    def _out_deal(self, deal):
        with open(join(
            self.dir_, self.prefix + str(deal) + self.suffix
        ), 'wt') as f:
            f.write(self.game.calc_deal_string(deal, self.rend))

    def run(self):
        if self.concat:
            idxs = self.idxs
            r = DealsRange(idxs=idxs)
            assert len(idxs) == 0
            with open(self.dir_, 'wt') as f:
                width = None
                for deal in r.run():
                    s = self.game.calc_deal_string(deal, self.rend)
                    w = len(s)
                    if width is None:
                        width = w
                    assert w == width
                    f.write(s)
            metadata_fn = self.dir_ + ".metadata.json"

            def i2s(i):
                assert isinstance(i, int)
                return "{}".format(i)
            which_deals = (
                "ms" if (self.which_deals == RandomBase.DEALS_MS)
                else "pysolfc"
            )
            metadata = {
                "game": self.game_variant,
                "seeds": [{
                    "end": i2s(r.end),
                    "start": i2s(r.start),
                    "type": "seq",
                }],
                "width": i2s(width),
                "which_deals": which_deals,
            }
            import json
            with open(metadata_fn, 'wt') as f:
                json.dump(metadata, f, sort_keys=True)

            return
        if self.mkdir_:
            dir_ = Path(self.dir_)
            if dir_.is_dir():
                if not self.force:
                    return
            os.makedirs(self.dir_, exist_ok=True)
        idxs = self.idxs
        while len(idxs):
            i = idxs[0]
            if i == 'seq':
                r = DealsRange(idxs=idxs)
                for deal in r.run():
                    self._out_deal(deal)
            elif i == 'slurp':
                idxs.pop(0)
                slurpfn = idxs.pop(0)
                with open(slurpfn, 'rt') as fh:
                    for line in fh:
                        self._out_deal(int(line))
            else:
                idxs.pop(0)
                self._out_deal(int(i))
        return 0


class ShlomifDwimGenDeals:
    def __init__(self):
        return

    def run(self):
        from pathlib import Path
        parent_dir = Path.home() / "Arcs" / "temp" / "solitaire" / \
            "deals" / "ms"
        os.makedirs(parent_dir, exist_ok=True)
        concat = True
        for game_variant in ["all_in_a_row", "black_hole", "freecell", "golf"]:
            dir_obj = parent_dir / game_variant
            dir_ = str(dir_obj)
            generator = GenMulti(
                concat=concat,
                dir_=dir_,
                force=False,
                game_variant=game_variant,
                idxs=["seq", "1", "32000",],
                mkdir_=True,
                ms=True,
                prefix="ms",
                suffix=".board",
            )
            generator.run()


def _cli_main():
    sys.exit(GenMultiParseCmdLine(argv=sys.argv).run())


if __name__ == "__main__":
    _cli_main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2019 Shlomi Fish <shlomif@cpan.org>
#
# Distributed under terms of the MIT license.

class SubclassResponsibility(Exception):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2019 Shlomi Fish <shlomif@cpan.org>
#
# Distributed under terms of the MIT license.

from pysol_cards.errors import SubclassResponsibility


class RandomBase(object):
    DEALS_PYSOL = 0
    DEALS_PYSOLFC = 1
    DEALS_MS = 2

    MAX_SEED = 10 ** 20
    ORIGIN_UNKNOWN = 0
    ORIGIN_RANDOM = 1
    ORIGIN_PREVIEW = 2
    ORIGIN_SELECTED = 3
    ORIGIN_NEXT_GAME = 4

    def __init__(self, seed=None):
        """docstring for __init__"""
        self.seed_as_string = None

    def shuffle(self, seq):
        for n in range(len(seq) - 1, 0, -1):
            j = self.randint(0, n)
            seq[n], seq[j] = seq[j], seq[n]
        return seq

    def randint(self, a, b):
        """ Get a random integer in the range [a, b] including both ends."""
        return a + int(self.random() * (b + 1 - a))

    def randrange(self, a, b):
        """ Get a random integer in the range [a, b) excluding b."""
        return self.randint(a, b - 1)

    def choice(self, sequence):
        """ Pick a random element of sequence """
        return sequence[self.randrange(0, len(sequence))]

    def setSeedAsStr(self, new_s):
        self.seed_as_string = new_s

    def getSeedAsStr(self):
        if self.seed_as_string:
            return self.seed_as_string
        return str(self)

    def getSeedStr(self):
        return str(self.initial_seed)

    def __str__(self):
        return self.str(self.initial_seed)

    def str(self, seed):
        return '%020d' % (seed)

    def increaseSeed(self, seed):
        if seed < self.MAX_SEED:
            return seed + 1
        return 0

    def copy(self):
        ret = self.__class__()
        ret.__dict__.update(self.__dict__)
        return ret

    def reset(self):
        raise SubclassResponsibility

    def _getRandomSeed(self):
        import time
        ret = int(time.time() * 256.0)
        return ((ret ^ (ret >> 24)) % (self.MAX_SEED + 1))

    def getstate(self):
        """getstate() for PySolFC"""
        return self.seed

    def setstate(self, new_state):
        """set to a new state"""
        self.seed = new_state



# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2019 Shlomi Fish <shlomif@cpan.org>
#
# Distributed under terms of the Expat license.

"""

"""

from pysol_cards.cards import Card
from pysol_cards.cards import createCards
from pysol_cards.random import shuffle


def empty_card():
    ret = Card(0, 0, 0)
    ret.empty = True
    return ret


class Columns(object):
    def __init__(self, num):
        self.cols = [[] for _ in range(num)]

    def add(self, idx, card):
        self.cols[idx].append(card)

    def rev(self):
        self.cols.reverse()


class Board(object):
    def __init__(self, num_columns, with_freecells=False,
                 with_talon=False, with_foundations=False):
        self.with_freecells = with_freecells
        self.with_talon = with_talon
        self.with_foundations = with_foundations
        self.raw_foundations_cb = None
        self.raw_foundations_line = None
        self.columns = Columns(num_columns)
        if self.with_freecells:
            self.freecells = []
        if self.with_talon:
            self.talon = []
        if self.with_foundations:
            self.foundations = [empty_card() for s in range(4)]
        self._lines = []

    def add_line(self, string):
        self._lines.append(string)

    def reverse_cols(self):
        self.columns.rev()

    def add(self, idx, card):
        self.columns.add(idx, card)

    def add_freecell(self, card):
        if not self.with_freecells:
            raise AttributeError("Layout does not have freecells!")
        self.freecells.append(card)

    def add_talon(self, card):
        if not self.with_talon:
            raise AttributeError("Layout does not have a talon!")
        self.talon.append(card)

    def put_into_founds(self, card):
        if not self.with_foundations:
            raise AttributeError("Layout does not have foundations!")
        res = self.foundations[card.suit].rank + 1 == card.rank
        if res:
            self.foundations[card.suit] = card
        return res

    def print_foundations(self, renderer):
        cells = []
        for f in [2, 0, 3, 1]:
            if not self.foundations[f].empty:
                cells.append(renderer.found_s(self.foundations[f]))

        if len(cells):
            self.add_line("Foundations:" + ("".join([" " + s for s in cells])))

    def gen_lines(self, renderer):
        self._lines = []
        if self.with_talon:
            self.add_line("Talon: " + renderer.l_concat(self.talon))
        if self.with_foundations:
            self.print_foundations(renderer)
        if self.raw_foundations_cb:
            self.add_line(self.raw_foundations_cb(renderer))
        elif self.raw_foundations_line:
            self.add_line(self.raw_foundations_line)
        if self.with_freecells:
            self.add_line("Freecells: " + renderer.l_concat(self.freecells))

        self._lines += [renderer.l_concat(c) for c in self.columns.cols]

    def calc_string(self, renderer):
        self.gen_lines(renderer)
        return "".join(line + "\n" for line in self._lines)


class Game(object):
    REVERSE_MAP = \
        {
            "freecell":
            ["forecell", "bakers_game",
             "ko_bakers_game", "kings_only_bakers_game",
             "relaxed_freecell", "eight_off"],
            "der_katz":
            ["der_katzenschwantz", "die_schlange"],
            "seahaven":
            ["seahaven_towers", "relaxed_seahaven",
             "relaxed_seahaven_towers"],
            "bakers_dozen": [],
            "gypsy": [],
            "klondike":
            ["klondike_by_threes",
             "casino_klondike", "small_harp", "thumb_and_pouch",
             "vegas_klondike", "whitehead"],
            "simple_simon": [],
            "yukon": [],
            "beleaguered_castle":
            ["streets_and_alleys", "citadel"],
            "fan": [],
            "black_hole": [],
            "all_in_a_row": [],
            "golf": [],
            "binary_star": [],
        }

    GAMES_MAP = {}
    for k, v in REVERSE_MAP.items():
        for name in [k] + v:
            GAMES_MAP[name] = k

    def __init__(self, game_id, game_num, which_deals, max_rank=13):
        self.game_id = game_id
        self.game_num = game_num
        self.which_deals = which_deals
        self.max_rank = max_rank
        self.game_class = self.GAMES_MAP[self.game_id]
        if not self.game_class:
            raise ValueError("Unknown game type " + self.game_id + "\n")

    def is_two_decks(self):
        return self.game_id in (
            "binary_star", "der_katz", "der_katzenschwantz",
            "die_schlange", "gypsy"
        )

    def get_num_decks(self):
        return 2 if self.is_two_decks() else 1

    def calc_deal_string(self, game_num, renderer):
        self.game_num = game_num
        self.deal()
        getattr(self, self.game_class)()
        return self.board.calc_string(renderer)

    def calc_layout_string(self, renderer):
        self.deal()
        getattr(self, self.game_class)()
        return self.board.calc_string(renderer)

    def print_layout(self, renderer):
        print(self.calc_layout_string(renderer), sep='', end='')

    def new_cards(self, cards):
        self.cards = cards
        self.card_idx = 0

    def deal(self):
        cards = shuffle(createCards(self.get_num_decks(),
                                    self.max_rank),
                        self.game_num, self.which_deals)
        cards.reverse()
        self.new_cards(cards)

    def __iter__(self):
        return self

    def no_more_cards(self):
        return self.card_idx >= len(self.cards)

    def __next__(self):
        if self.no_more_cards():
            raise StopIteration
        c = self.cards[self.card_idx]
        self.card_idx += 1
        return c

    def next(self):
        return self.__next__()

    def add(self, idx, card):
        self.board.add(idx, card)

    def add_freecell(self, card):
        self.board.add_freecell(card)

    def cyclical_deal(self, num_cards, num_cols, flipped=False):
        for i in range(num_cards):
            self.add(i % num_cols, next(self).flip(flipped=flipped))

    def add_all_to_talon(self):
        for c in self:
            self.board.add_talon(c)

    def add_empty_fc(self):
        self.add_freecell(empty_card())

    def _shuffleHookMoveSorter(self, cards, cb, ncards):
        extracted, i, new = [], len(cards), []
        for c in cards:
            select, ord_ = cb(c)
            if select:
                extracted.append((ord_, i, c))
                if len(extracted) >= ncards:
                    new += cards[(len(cards) - i + 1):]
                    break
            else:
                new.append(c)
            i -= 1
        return new, [x[2] for x in reversed(sorted(extracted))]

    def _shuffleHookMoveToBottom(self, inp, cb, ncards=999999):
        cards, scards = self._shuffleHookMoveSorter(inp, cb, ncards)
        return scards + cards

    def _shuffleHookMoveToTop(self, inp, cb, ncards=999999):
        cards, scards = self._shuffleHookMoveSorter(inp, cb, ncards)
        return cards + scards

    def all_in_a_row(game):
        game.board = Board(13)
        game.cards = game._shuffleHookMoveToTop(
            game.cards,
            lambda c: (c.id == 13, c.suit),
            1)
        game.cyclical_deal(52, 13)
        game.board.raw_foundations_line = 'Foundations: -'

    def bakers_dozen(game):
        n = 13
        cards = list(reversed(game.cards))
        for i in [i for i, c in enumerate(cards) if c.is_king()]:
            j = i % n
            while j < i:
                if not cards[j].is_king():
                    cards[i], cards[j] = cards[j], cards[i]
                    break
                j += n
        game.new_cards(cards)
        game.board = Board(13)
        game.cyclical_deal(52, 13)

    def beleaguered_castle(game):
        game.board = Board(8, with_foundations=True)
        if game.game_id in ('beleaguered_castle', 'citadel'):
            new = []
            for c in game:
                if c.is_ace():
                    game.board.put_into_founds(c)
                else:
                    new.append(c)
            game.new_cards(new)
        for _ in range(6):
            for s in range(8):
                c = next(game)
                cond1 = game.game_id == 'citadel'
                if not (cond1 and game.board.put_into_founds(c)):
                    game.add(s, c)
            if game.no_more_cards():
                break
        if game.game_id == 'streets_and_alleys':
            game.cyclical_deal(4, 4)

    def _black_hole_generic(game, num_decks, excludes, f):
        game.board = Board(17)
        game.cards = game._shuffleHookMoveToBottom(
            game.cards,
            lambda c: (c.id in excludes, c.suit),
            2)
        for _ in range(num_decks):
            next(game)
        game.cyclical_deal(num_decks * (52 - 1), 17)
        game.board.raw_foundations_line = 'Foundations: ' + f

    def binary_star(game):
        return game._black_hole_generic(
            num_decks=2, excludes=frozenset((13, 38)), f='AS KH')

    def black_hole(game):
        return game._black_hole_generic(
            num_decks=1, excludes=frozenset((13, )), f='AS')

    def der_katz(game):
        is_ds = game.game_id == 'die_schlange'
        if is_ds:
            print('Foundations: H-A S-A D-A C-A H-A S-A D-A C-A')
        game.board = Board(9)
        i = 0
        for c in game:
            if c.is_king():
                i += 1
            if not (is_ds and c.is_ace()):
                game.add(i, c)

    def fan(game):
        game.board = Board(18)
        game.cyclical_deal(52 - 1, 17)
        game.add(17, next(game))

    def freecell(game):
        is_fc = (game.game_id in ("forecell", "eight_off"))
        game.board = Board(8, with_freecells=is_fc)
        max_rank = (game.max_rank - 1 if is_fc else game.max_rank)
        game.cyclical_deal(4 * max_rank, 8)

        if is_fc:
            for c in game:
                game.add_freecell(c)
                if game.game_id == "eight_off":
                    game.add_empty_fc()

    def golf(game):
        num_cols = 7
        game.board = Board(num_cols, with_talon=True)
        game.cyclical_deal(num_cols * 5, num_cols)
        game.add_all_to_talon()
        card = game.board.talon.pop(0)
        game.board.raw_foundations_cb = lambda renderer: 'Foundations: ' + \
            renderer.l_concat([card])

    def gypsy(game):
        num_cols = 8
        game.board = Board(num_cols, with_talon=True)
        game.cyclical_deal(num_cols * 2, num_cols, flipped=True)
        game.cyclical_deal(num_cols, num_cols)
        game.add_all_to_talon()

    def klondike(game):
        num_cols = 7
        game.board = Board(num_cols, with_talon=True)
        for r in range(num_cols - 1, 0, -1):
            game.cyclical_deal(r, r, flipped=True)
        game.cyclical_deal(num_cols, num_cols)
        game.add_all_to_talon()
        if not (game.game_id == 'small_harp'):
            game.board.reverse_cols()

    def seahaven(game):
        game.board = Board(10, with_freecells=True)
        game.cyclical_deal(50, 10)
        game.add_empty_fc()
        for c in game:
            game.add_freecell(c)

    def simple_simon(game):
        game.board = Board(10)
        for num_cards in range(9, 2, -1):
            game.cyclical_deal(num_cards, num_cards)
        game.cyclical_deal(10, 10)

    def yukon(game):
        num_cols = 7
        game.board = Board(num_cols)
        for i in range(1, num_cols):
            for j in range(i, num_cols):
                game.add(j, next(game).flip())
        for i in range(4):
            for j in range(1, num_cols):
                game.add(j, next(game))
        game.cyclical_deal(num_cols, num_cols)



# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2020 Shlomi Fish <shlomif@cpan.org>
#
# Distributed under terms of the MIT license.

"""

"""

from pysol_cards.random import RandomBase
from pysol_cards.random import match_ms_deal_prefix


class SingleDealArgsParser(object):
    def __init__(self, args):
        self.print_ts = False
        self.which_deals = RandomBase.DEALS_PYSOL
        self.max_rank = 13
        while args[1][0] == '-':
            a = args[1]
            args.pop(1)
            if a == "-t":
                self.print_ts = True
            elif (a == "--max-rank"):
                self.max_rank = int(args.pop(1))
            elif (a == "--pysolfc") or (a == "-F"):
                self.which_deals = RandomBase.DEALS_PYSOLFC
            elif (a == "--ms") or (a == "-M"):
                self.which_deals = RandomBase.DEALS_MS
            else:
                raise ValueError("Unknown flag " + a + "!")

        game_num_s = args[1]
        msgame = match_ms_deal_prefix(game_num_s)
        if msgame is not None:
            if self.which_deals == RandomBase.DEALS_MS or \
                    self.which_deals == RandomBase.DEALS_PYSOL:
                self.which_deals = RandomBase.DEALS_MS
                self.game_num = msgame
            else:
                raise ValueError("ms deals mismatch")
        else:
            self.game_num = int(game_num_s)
        self.which_game = args[2] if len(args) >= 3 else "freecell"
